/*
 * mvs-minfo-provider.c
 *
 * This file is part of maevies
 * Copyright (C) 2010 Simón Pena <spenap@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 */

#include "mvs-minfo-provider.h"

#include <libxml/parser.h>
#include <libxml/xpath.h>

#include "mvs-tmdb-movie.h"

#define TMDB_API_KEY "249e1a42df9bee09fac5e92d3a51396b"
#define TMDB_LANGUAGE "en"
#define TMDB_FORMAT "xml"
#define TMDB_METHOD "Movie.search"
#define TMDB_BASE_URL "http://api.themoviedb.org/2.1/%s/%s/%s/%s/%s"
#define TMDB_MOVIE_XPATH "/OpenSearchDescription/movies/movie"

G_DEFINE_TYPE (MvsMInfoProvider, mvs_minfo_provider, G_TYPE_OBJECT)

enum {
        PROP_0,
        PROP_FORMAT,
};

#define GET_PRIVATE(o) \
  (G_TYPE_INSTANCE_GET_PRIVATE ((o), MVS_TYPE_MINFO_PROVIDER, MvsMInfoProviderPrivate))

struct _MvsMInfoProviderPrivate {
        gchar *format;
};

enum {
        RESPONSE_RECEIVED,
        LAST_SIGNAL
};

static guint
signals[LAST_SIGNAL] = { 0 };


static void
mvs_minfo_provider_get_property (GObject *object, guint property_id,
                         GValue *value, GParamSpec *pspec)
{
        MvsMInfoProvider *self = MVS_MINFO_PROVIDER (object);

        switch (property_id) {
        case PROP_FORMAT:
                g_value_set_string (value, self->priv->format);
                break;
        default:
                G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
        }
}

static void
mvs_minfo_provider_set_property (GObject *object, guint property_id,
                         const GValue *value, GParamSpec *pspec)
{
        MvsMInfoProvider *self = MVS_MINFO_PROVIDER (object);

        switch (property_id) {
        case PROP_FORMAT:
                mvs_minfo_provider_set_format (self,
                                g_value_get_string (value));
                break;
        default:
                G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
        }
}

static void
mvs_minfo_provider_finalize (GObject *object)
{
        MvsMInfoProvider *self = MVS_MINFO_PROVIDER (object);

        g_free (self->priv->format);

        G_OBJECT_CLASS (mvs_minfo_provider_parent_class)->finalize (object);
}

static void
mvs_minfo_provider_class_init (MvsMInfoProviderClass *klass)
{
        GObjectClass *object_class = G_OBJECT_CLASS (klass);

        g_type_class_add_private (klass, sizeof (MvsMInfoProviderPrivate));

        object_class->get_property = mvs_minfo_provider_get_property;
        object_class->set_property = mvs_minfo_provider_set_property;
        object_class->finalize = mvs_minfo_provider_finalize;

        g_object_class_install_property
                (object_class, PROP_FORMAT,
                 g_param_spec_string ("format", "The format", "The format",
                                      TMDB_FORMAT,
                                      G_PARAM_READWRITE | G_PARAM_CONSTRUCT));

        signals[RESPONSE_RECEIVED] = g_signal_new ("response-received", MVS_TYPE_MINFO_PROVIDER,
                        G_SIGNAL_RUN_LAST | G_SIGNAL_NO_RECURSE | G_SIGNAL_NO_HOOKS,
                        G_STRUCT_OFFSET (MvsMInfoProviderClass, response_callback),
                        NULL,
                        NULL,
                        g_cclosure_marshal_VOID__POINTER,
                        G_TYPE_NONE,
                        1,
                        G_TYPE_POINTER);
}

static void
mvs_minfo_provider_init (MvsMInfoProvider *self)
{
        self->priv = GET_PRIVATE (self);
        self->priv->format = NULL;
}

MvsMInfoProvider*
mvs_minfo_provider_new (void)
{
        return g_object_new (MVS_TYPE_MINFO_PROVIDER, NULL);
}

static MvsTmdbMovie*
generate_movie_info (xmlNodePtr node)
{
        xmlNodePtr cur_node = NULL;
        MvsTmdbMovie *movie_info = mvs_tmdb_movie_new ();

        for (cur_node = node; cur_node; cur_node = cur_node->next) {
                if (cur_node->type == XML_ELEMENT_NODE) {
                        gchar *value = xmlNodeGetContent (cur_node);

                        g_object_set (movie_info, cur_node->name, value, NULL);
                }
        }
        return movie_info;
}

static GList*
generate_list (xmlNodeSetPtr node_set)
{
        int i = 0;
        GList *list = NULL;

        for (i = 0; i < node_set->nodeNr; i++) {
                xmlNodePtr node = node_set->nodeTab[i];
                if (node->type == XML_ELEMENT_NODE) {
                        MvsTmdbMovie *movie_info =
                                        generate_movie_info (node->children);
                        if (movie_info)
                                list = g_list_prepend (list, movie_info);
                }
        }

        if (list)
                list = g_list_reverse (list);

        return list;
}

static GList*
parse_xml (const char *xml_data, goffset length)
{
        GList *list = NULL;
        xmlDocPtr document = xmlReadMemory (xml_data, length,
                        NULL,
                        NULL,
                        XML_PARSE_NOBLANKS | XML_PARSE_RECOVER);
        g_return_if_fail (document);

        xmlXPathContextPtr context_ptr = xmlXPathNewContext (document);

        xmlXPathObjectPtr xpath_obj =
                        xmlXPathEvalExpression (TMDB_MOVIE_XPATH, context_ptr);

        xmlNodeSetPtr nodeset = xpath_obj->nodesetval;

        if (nodeset->nodeNr > 0) {
                list = generate_list (nodeset);
        }

        xmlXPathFreeNodeSetList (xpath_obj);
        xmlXPathFreeContext (context_ptr);
        xmlFreeDoc (document);

        return list;
}

static void
process_response (SoupSession *session, SoupMessage *message,
                    gpointer user_data)
{
        MvsMInfoProvider *self = MVS_MINFO_PROVIDER (user_data);
        const gchar *mime = NULL;
        GList *list = NULL;

        if (!SOUP_STATUS_IS_SUCCESSFUL (message->status_code) ||
                        message->response_body->length <= 0) {

                g_print ("%s\n", message->reason_phrase);
        }
        else {

                mime = soup_message_headers_get_content_type
                                (message->response_headers, NULL);

                list = parse_xml (message->response_body->data,
                                message->response_body->length);
        }

        g_signal_emit (self, signals[RESPONSE_RECEIVED], 0, list);
}

static gchar *
get_query_uri (MvsMInfoProvider *self, const char *query)
{
        /* METHOD/LANGUAGE/FORMAT/APIKEY/MOVIENAME */
        gchar *uri = g_strdup_printf (TMDB_BASE_URL, TMDB_METHOD,
                        TMDB_LANGUAGE,
                        self->priv->format,
                        TMDB_API_KEY,
                        query);

        return uri;
}

gboolean
mvs_minfo_provider_query (MvsMInfoProvider *self,
                          const gchar *query)
{
        g_return_val_if_fail (MVS_IS_MINFO_PROVIDER (self), FALSE);

        SoupSession *session = NULL;
        SoupMessage *message = NULL;
        gboolean message_queued = FALSE;

        gchar *uri = get_query_uri (self, query);

        g_return_val_if_fail (uri, FALSE);

        session = soup_session_async_new ();
        message = soup_message_new ("GET", uri);

        if (message) {
                soup_session_queue_message (session, message,
                                process_response, self);
                message_queued = TRUE;
        }

        g_free (uri);

        return message_queued;
}

gboolean
mvs_minfo_provider_set_format (MvsMInfoProvider *self,
                               const gchar *format)
{
        g_return_val_if_fail (MVS_IS_MINFO_PROVIDER (self), FALSE);

        g_free (self->priv->format);

        self->priv->format = g_strdup (format);

        return TRUE;
}
